import enum
from typing import Generic, Optional, Sequence, Any, TypeVar
import datetime
import pendulum
from typing_extensions import Annotated
from pydantic import BaseModel, ConfigDict, PlainValidator, PlainSerializer, Field
from pydantic.alias_generators import to_camel


_TZ = pendulum.timezone("Europe/Rome")
_DATE_FMT = "%Y-%m-%d"


def validate_date(val: Any):
    if val is None:
        return None
    dt = pendulum.DateTime.strptime(val, _DATE_FMT).astimezone(_TZ)
    return dt


def serialize_date(dt: pendulum.DateTime | None):
    if dt is None:
        return None
    return dt.strftime(_DATE_FMT)


Date = Annotated[
    pendulum.DateTime,
    PlainValidator(validate_date),
    PlainSerializer(serialize_date),
]


def validate_datetime(val: Any):
    if val is None:
        return None

    if isinstance(val, float) or isinstance(val, int):
        return pendulum.DateTime.fromtimestamp(1e-3 * val, tz=_TZ)

    raise ValueError("Invalid date type")


def serialize_datetime(dt: pendulum.DateTime | None):
    if dt is None:
        return None
    return dt.timestamp()


DateTime = Annotated[
    pendulum.DateTime,
    PlainSerializer(serialize_datetime),
    PlainValidator(validate_datetime),
]


def validate_minutes_delta(td: Any):
    if td is None:
        return None

    if isinstance(td, float) or isinstance(td, int):
        return datetime.timedelta(minutes=td)

    raise ValueError("Invalid data type")


def serialize_minutes_delta(td: datetime.timedelta):
    return td.seconds // 60


MinutesDelta = Annotated[
    datetime.timedelta,
    PlainValidator(validate_minutes_delta),
    PlainSerializer(serialize_minutes_delta),
]


def validate_id_list(l: Any):
    if not isinstance(l, str):
        raise ValueError("Value must be a str")

    return [int(e) for e in l.split(",")]


def serialize_id_list(l: Sequence[int]):
    return ",".join([str(e) for e in l])


IdList = Annotated[
    Sequence[int], PlainValidator(validate_id_list), PlainSerializer(serialize_id_list)
]


class ApiModel(BaseModel):
    """Base class for an API model"""

    model_config = ConfigDict(
        use_enum_values=True,
        alias_generator=to_camel,
    )


class ApiError(ApiModel):
    """Model of an API error"""

    key: str
    """Unique error key which defines specific case"""
    message: str
    """Human-readable error message"""
    stack_trace: Optional[str] = None
    """Stack-trace info for developers"""
    fields: Optional[str] = None
    """Field or fields (if any) related to error"""


A = TypeVar("A", bound=ApiModel)


class ItemList(ApiModel, Generic[A]):
    """Base list of items"""

    items: Sequence[A]
    """List of items matching criteria"""
    offset: int
    """Offset position"""
    limit: int
    """Max number of returned items"""


class AllowedAction(ApiModel):

    can_modify: Optional[bool] = None
    can_delete: Optional[bool] = None


class WithAllowedAction(ApiModel):

    allowed_actions: AllowedAction


class WithIdModel(ApiModel):
    uid: int = Field(alias="id")
    """Unique object identifier"""


class Customer(WithIdModel):

    name: str
    """Customer name"""
    archived: bool
    """True if the customer is archived"""
    created: DateTime
    """Creation date and time"""
    url: str
    """URL of the customer"""
    description: Optional[str] = None
    """Customer description"""


class CustomerWithAllowedAction(Customer, WithAllowedAction):
    pass


class CustomerList(ItemList[CustomerWithAllowedAction]):
    pass


class User(WithIdModel):
    department_id: int
    """Unique identifier of user department"""
    time_zone_group_id: int
    """Unique identifier of user time zone group"""
    hired: Optional[Date] = None
    """User's hire date"""
    release_date: Optional[Date] = None
    """User's release date"""
    email: Optional[str] = None
    """User email"""
    full_name: str
    """Fullname: first_name + middle_name + last_name"""
    username: Optional[str] = None
    """Unique username"""
    active: bool
    """True if the user is active"""
    first_name: str
    """User's first name"""
    middle_name: str
    """User's middle name"""
    last_name: str
    """User's last name"""


class UserAllowedActions(ApiModel):

    can_submit_timetrack: bool


class UserWithAllowedActions(User):

    allowed_actions: UserAllowedActions


class UserList(ItemList[UserWithAllowedActions]):
    departments: Optional[dict[int, "Department"]] = None
    """Map of user department id to department object. Contains departments referenced in users referenced in data list"""
    time_zone_groups: Optional[dict[int, "TimeZoneGroup"]] = None
    """Map of user time zone id to user time zone object. Contains user time zone referenced in users referenced in data list"""


class UserSchedule(ApiModel):

    date_from: Date
    """Date the schedule returned from, inclusive."""
    date_to: Date
    """Date the schedule returned to, inclusive."""
    schedule: Sequence[MinutesDelta]
    """Array of workday durations, where the first element corresponds to 'dateFrom' and the last to 'dateTo'."""


class Comment(WithIdModel):
    user_id: int
    """Unique user identifier"""
    created: DateTime
    """Creation date and time"""
    updated: DateTime
    """Date and time of last editing"""
    updating_user_id: int
    """Unique identifier of the user who made the last editing"""
    text: str
    """Text of the comment"""


class CommentList(ItemList[Comment]):
    users: Optional[dict[int, User]] = None
    """Map of userId to user"""


class Department(WithIdModel):
    name: str
    """Department name"""
    default: bool
    """True if this is the default department"""


class DepartmentList(ItemList[Department]):
    pass


class EntityFilterBean(ApiModel):
    customer_ids: IdList
    project_ids: IdList
    task_ids: IdList
    user_ids: IdList
    department_ids: IdList


class Event(str, enum.Enum):
    create = "create"
    update = "update"
    move = "move"
    delete = "delete"
    task = "task"
    task_create = "task.create"
    task_update = "task.update"
    task_move = "task.move"
    task_delete = "task.delete"
    project = "project"
    project_create = "project.create"
    project_update = "project.update"
    project_move = "project.move"
    project_delete = "project.delete"
    customer = "customer"
    customer_create = "customer.create"
    customer_update = "customer.update"
    customer_delete = "customer.delete"
    department = "department"
    department_create = "department.create"
    department_update = "department.update"
    department_delete = "department.delete"
    user = "user"
    user_create = "user.create"
    user_update = "user.update"
    user_delete = "user.delete"
    time_zone_group = "timeZoneGroup"
    time_zone_group_create = "timeZoneGroup.create"
    time_zone_group_update = "timeZoneGroup.update"
    time_zone_group_delete = "timeZoneGroup.delete"
    leave_time = "leaveTime"
    leave_time_update = "leaveTime.update"
    timetrack = "timetrack"
    timetrack_update = "timetrack.update"
    timetrack_approval_status = "timetrackApprovalStatus"
    timetrack_approval_status_update = "timetrackApprovalStatus.update"


class WithDetails(WithIdModel):
    event: Event
    """Which events shall be sent to specified URL"""
    entity_filter: EntityFilterBean = Field(alias="filter")
    """"""
    enabled: bool
    """Is this subscription active or not"""
    lastHttpStatus: int
    """HTTP Status of last call to target url"""
    lastHttpError: str
    """HTTP error of last call to target url"""
    lastHttpCall: DateTime
    """Date of last call to target url"""
    user_id: int
    """Id of creator"""
    target_url: str
    """URL which will be used to send events"""


class Company(ApiModel):
    logo_uri: str
    """Url of logo used in web interface"""
    name: str
    """Company name"""


class CustomName(ApiModel):
    singular: str
    plural: str


class CustomNames(ApiModel):
    first_level: CustomName
    second_level: CustomName
    third_level: CustomName
    department: CustomName


class TimeFormat(str, enum.Enum):
    minutes = "minutes"
    hours = "hours"


class ClockFormat(str, enum.Enum):
    clock_format_12 = "clock_format_12"
    clock_format_24 = "clock_format_24"


class Format(ApiModel):
    currency: str
    """Which currency character is set in system"""
    decimal_separator: str
    """Which decimal separator is set in system"""
    time_format: TimeFormat
    """Which time format is preferred in reports.
    If 'minutes', then HH:MM (i.e. 03:45); if 'hours', then HH. (i.e. 03.75)"""
    clock_format: ClockFormat
    """Which clock format is preferred.
    If '12-hour clock', then (1-12):MM AM/PM (i.e. 03:45 AM); if 
    '24-hors clock', then (1-24):MM (i.e. 15:45)"""
    day_of_week_start: int
    """On which day week starts. 
    Day number corresponds Javascript, i.e. 0 = sunday, 1 = monday, 6 = saturday"""


class Features(ApiModel):
    overtime_registration: bool
    """Overtime Registration feature (is time exceeding the work day duration 
    reported as overtime)"""
    departments: bool
    """Departments feature (are users grouped by departments)"""
    time_zone_groups: bool
    """Time Zone Groups feature (are users grouped by time zone groups)"""
    task_estimates: bool
    """Task Estimates feature (do tasks contain estimated time)"""
    task_workflow: bool
    """Task Workflow feature (is there are many workflow statuses or tasks have 
    only simple Open/Completed statuses)"""
    work_assignments: bool
    """Work Assignments features (is it possible to assign specific users to 
    specific tasks)"""
    types_of_work: bool
    """Types of Work feature (do tasks have assigned types of works or not)"""
    leavetime_registration: bool
    """Leave Time Tracking feature (is leave time also registered in time-sheet 
    or not)"""


class Urls(ApiModel):
    actitime: str = Field(alias="actiTime")
    """Home of actiTIME web application"""
    actiplans: Optional[str] = Field(default=None, alias="actiPlans")
    """Home of actiPLANS web application"""
    api: str
    """Base uri of API"""
    api_documentation: str
    """Uri of API documentation"""


class Limits(ApiModel):
    max_batch_size: int
    """How much requests could be sent via single batch request"""
    max_query_limit: int
    """How much results could be returned in single listing request"""


class Info(ApiModel):

    company: Company
    custom_names: CustomNames
    local_format: Format = Field(alias="format")
    features: Features
    """Describes which features are enabled in actiTIME Web UI. Note that API 
    endpoint is available even if corresponding feature is turned off"""
    urls: Urls
    limits: Limits
    server_uuid: str = Field(alias="serverUUID")


class TimeZoneGroup(WithIdModel):
    """Represents single location with associated time zone where users could
    be assigned"""

    name: str
    """Human-readable name of associated location"""
    time_zone_id: str
    """Time Zone Id in format of IANA Time Zone Database"""
    default: bool
    """True if this is default time zone group"""


class LeaveTypeBalance(str, enum.Enum):
    none = "None"
    sick = "Sick"
    pto = "PTO"


class LeaveType(WithIdModel):
    """Represents leave type object"""

    name: str
    """Name of leave type"""
    balance: LeaveTypeBalance
    """Which balance is affected by leave of this type"""
    archived: bool
    """True if this leave type is archived"""


class LeaveTimeRecord(ApiModel):
    """Contains time leaved and type of leave for specific user"""

    user_id: int
    """Unique user identifier"""
    day_offset: int
    """Offset for 'dateFrom'"""
    date: Date
    """Leave date"""
    leave_type_id: int
    """Leave type identifier"""
    leave_time: MinutesDelta
    """Duration of leave time"""


class LeaveTimeList(ApiModel):
    users: Optional[dict[int, User]] = None
    """Map of user id to user. Contains users referenced in data list."""
    departments: Optional[dict[int, Department]] = None
    """Map of user department id to department object. 
    Contains departments referenced in users referenced in data list."""
    time_zone_groups: Optional[dict[int, TimeZoneGroup]] = None
    date_from: Date
    """Start date of returned data"""
    date_to: Date
    """End date of returned data"""
    next_date_from: Optional[Date] = None
    """If not null, then only partial results were included to satisfy 
    'stopAfter' parameter - till date specified in 'dateTo'. To request more 
    you need to repeat your request from date specified in 'nextDateFrom' field.
    """
    leave_types: Optional[dict[int, LeaveType]] = None
    """Map of id to leave type. Contains leave types referenced in data list"""
    data: Sequence[LeaveTimeRecord]
    """List of user leave time records"""


class LeaveTypeList(ItemList[LeaveType]):
    pass


class Project(WithIdModel):

    customer_id: int
    name: str
    archived: bool
    created: DateTime
    url: str
    customer_name: str
    description: str | None = None


class ProjectWithAllowedActions(Project, WithAllowedAction):
    pass


class ProjectList(ItemList[ProjectWithAllowedActions]):

    customers: Optional[dict[int, Customer]] = None


class TypeOfWork(WithIdModel):
    name: str
    rate: Optional[float]
    archived: bool
    billable: bool
    default: bool


class WorkflowTypeEnum(str, enum.Enum):
    open = "open"
    completed = "completed"


class WorkflowStatus(WithIdModel):
    name: str
    type_: WorkflowTypeEnum = Field(alias="type")


class WorkflowStatusWithAllowedActions(WorkflowStatus, AllowedAction):
    pass


class WorkflowStatusList(ItemList[WorkflowStatus]):
    pass


class TaskStatus(str, enum.Enum):

    open_ = "open"
    completed = "completed"


class Task(WithIdModel):

    name: str
    description: Optional[str] = None
    created: DateTime
    status: TaskStatus
    workflow_status_id: int
    type_of_work_id: int
    url: str
    project_name: str
    customer_name: str
    workflow_status_name: str
    type_of_work_name: str
    deadline: Optional[DateTime] = None
    estimated_time: Optional[MinutesDelta] = None
    customer_id: int
    project_id: int


class TaskWithAllowedActions(Task, WithAllowedAction):
    pass


class TaskList(ItemList[TaskWithAllowedActions]):

    customers: Optional[dict[int, Customer]] = None
    projects: Optional[dict[int, Project]] = None
    types_of_work: Optional[dict[int, TypeOfWork]] = None
    workflow_statuses: Optional[dict[int, WorkflowStatus]] = None


class TimeTrackRecord(ApiModel):
    """Contains time tracked for specific task and comment"""

    task_id: int
    """Identifier of task the time is tracked for"""
    time: MinutesDelta
    """Tracked time"""
    comment: Optional[str] = None
    """Optional comment for time-track record"""


class UserDayTTDataWithDayOffset(ApiModel):
    """Contains tracked time for specified user and date. Date is defined as offset from 'dateFrom' field of wrapping object"""

    user_id: int
    """Identifier of user that tracked time"""
    records: list[TimeTrackRecord]
    """List of time-track records for day"""
    approved: Optional[bool] = None
    """Approval status of the day"""
    day_offset: int
    """Offset in days from 'dateFrom' field of wrapping object"""
    date: Date
    """Time track date"""


class TimeTrackList(ApiModel):
    """Result of time track request"""

    users: Optional[dict[int, User]] = None
    """Map of user id to user object. Contains users referenced in data list"""
    departments: Optional[dict[int, Department]] = None
    """Map of user department id to department object. Contains departments referenced in users referenced in data list"""
    time_zone_group: Optional[dict[int, TimeZoneGroup]] = None
    """Map of user time zone id to user time zone object. Contains user time zone referenced in users referenced in data list"""
    date_from: Date
    """Start date of returned data"""
    date_to: Date
    """End date of returned data. Note that this one may be smaller than requested in parameters"""
    next_date_from: Optional[Date] = None
    """If not null, then only partial results were included to satisfy 
    'stopAfter' parameter - till date specified in 'dateTo'. To request more 
    you need to repeat your request from date specified in 'nextDateFrom' 
    field."""
    customers: Optional[dict[int, Customer]] = None
    """Map of customer id to customer object. Contains customers referenced in data list"""
    projects: Optional[dict[int, Project]] = None
    """Map of project id to project object. Contains projects referenced in data list"""
    tasks: Optional[dict[int, Task]] = None
    """Map of task id to task object. Contains tasks referenced in data list"""
    types_of_work: Optional[dict[int, TypeOfWork]] = None
    """Map of type of work id to type of work object. Contains types of work referenced in tasks referenced in data list"""
    workflow_statuses: Optional[dict[int, WorkflowStatus]] = None
    """Map of workflow status id to workflow status object. Contains workflow status objects referenced in tasks referenced in data list"""
    data: list[UserDayTTDataWithDayOffset]
    """List of time track records per user per day"""


class TimeZoneGroupList(ItemList[TimeZoneGroup]):
    pass


class TypeOfWorkList(ItemList[TypeOfWork]):
    pass



from typing import TypedDict, NotRequired, Sequence, Literal
from pendulum import DateTime


class ParamsDict(TypedDict):
    pass


class GetParamsDict(ParamsDict):

    offset: NotRequired[int]
    limit: NotRequired[int]


class GetParamsWithNameWords(GetParamsDict):
    name: NotRequired[str]
    words: NotRequired[str]


class GetCommentsParams(GetParamsDict):

    include_referenced: NotRequired[Sequence[Literal["users"]]]


class GetCustomersParams(GetParamsWithNameWords):

    ids: NotRequired[Sequence[int]]
    sort: NotRequired[Literal["+created", "-created", "+name", "-name"]]
    archived: NotRequired[bool]


class GetDepartmentsParams(GetParamsDict):
    name: NotRequired[str]
    sort: NotRequired[Literal["+name", "-name"]]


class GetLeavetimeParams(ParamsDict):

    user_ids: NotRequired[Sequence[int]]
    leave_type_ids: NotRequired[Sequence[int]]
    date_from: DateTime
    date_to: NotRequired[DateTime]
    stop_after: NotRequired[int]
    include_referenced: NotRequired[
        Sequence[Literal["departments", "timeZoneGroups", "users", "leaveTypes"]]
    ]


class GetLeaveTypeParams(GetParamsWithNameWords):

    ids: NotRequired[Sequence[int]]
    balance: NotRequired[Literal["None", "Sick", "PTO"]]
    archived: NotRequired[bool]
    sort: NotRequired[Literal["+name", "-name"]]


class GetProjectsParams(GetParamsWithNameWords):

    ids: NotRequired[Sequence[int]]
    customer_ids: NotRequired[Sequence[int]]
    sort: NotRequired[Literal["+created", "-created", "+name", "-name"]]
    archived: NotRequired[bool]
    include_referenced: NotRequired[Sequence[Literal["customers"]]]


class GetTasksParams(GetParamsWithNameWords):
    ids: NotRequired[Sequence[int]]
    customer_ids: NotRequired[Sequence[int]]
    project_ids: NotRequired[Sequence[int]]
    type_of_work_ids: NotRequired[Sequence[int]]
    workflow_status_ids: NotRequired[Sequence[int]]
    sort: NotRequired[
        Literal["+created", "-created", "+name", "-name", "+status", "-status"]
    ]
    status: NotRequired[Literal["open", "completed"]]
    include_referenced: NotRequired[
        Sequence[Literal["customers", "typesOfWork", "projects", "workflowStatuses"]]
    ]


class GetTimetrackParams(ParamsDict):
    user_ids: NotRequired[Sequence[int]]
    task_ids: NotRequired[Sequence[int]]
    project_ids: NotRequired[Sequence[int]]
    customer_ids: NotRequired[Sequence[int]]
    approved: NotRequired[bool]
    date_from: DateTime
    date_to: NotRequired[DateTime]
    stop_after: NotRequired[int]
    include_referenced: NotRequired[
        Sequence[
            Literal[
                "departments",
                "typesOfWork",
                "timeZoneGroups",
                "comments",
                "users",
                "tasks",
                "customers",
                "approvalStatus",
                "projects",
                "workflowStatuses",
            ]
        ]
    ]


class GetTimeZoneGroupsParams(GetParamsDict):
    name: NotRequired[str]
    sort: NotRequired[Literal["+name", "-name"]]


class GetTypesOfWorkParams(GetParamsWithNameWords):

    ids: NotRequired[Sequence[int]]
    archived: NotRequired[bool]
    billable: NotRequired[bool]
    sort: NotRequired[Literal["+name", "-name"]]


class GetUsersParams(GetParamsDict):
    email: NotRequired[str]
    ids: NotRequired[Sequence[int]]
    department: NotRequired[Sequence[int]]
    time_zone_group: NotRequired[Sequence[int]]
    active: NotRequired[bool]
    username: NotRequired[str]
    name: NotRequired[str]
    sort: NotRequired[
        Literal[
            "+lastName",
            "-lastName",
            "+hired",
            "-hired",
            "+department",
            "-department",
            "+timeZoneGroup",
            "-timeZoneGroup",
            "+username",
            "-username",
        ]
    ]
    include_referenced: NotRequired[Sequence[Literal["departments", "timeZoneGroups"]]]


class GetUserScheduleParams(ParamsDict):
    date_from: DateTime
    date_to: NotRequired[DateTime]


class GetWorkflowStatusesParams(GetParamsWithNameWords):
    ids: NotRequired[Sequence[int]]
    type: NotRequired[Literal["open", "completed"]]
    sort: NotRequired[Literal["+name", "-name", "+type", "-type"]]



from enum import Enum
from pendulum import DateTime
from typing import (
    Optional,
    Sequence,
    Type,
    TypeVar,
    Unpack,
)
import requests
from requests.auth import AuthBase
from pydantic.alias_generators import to_camel
from .models import (
    serialize_date,
    ApiModel,
    ApiError,
    CustomerWithAllowedAction,
    CustomerList,
    CommentList,
    Department,
    DepartmentList,
    WithDetails,
    Info,
    LeaveTimeList,
    LeaveType,
    LeaveTypeList,
    ProjectWithAllowedActions,
    ProjectList,
    TaskWithAllowedActions,
    TaskList,
    TimeTrackRecord,
    TimeTrackList,
    TimeZoneGroup,
    TimeZoneGroupList,
    TypeOfWork,
    TypeOfWorkList,
    UserWithAllowedActions,
    UserList,
    UserSchedule,
    WorkflowStatusWithAllowedActions,
    WorkflowStatusList,
)
from ._params import (
    ParamsDict as _ParamsDict,
    GetCustomersParams as _GetCustomersParams,
    GetCommentsParams as _GetCommentsParams,
    GetDepartmentsParams as _GetDepartmentsParams,
    GetLeavetimeParams as _GetLeavetimeParams,
    GetLeaveTypeParams as _GetLeaveTypeParams,
    GetProjectsParams as _GetProjectsParams,
    GetTasksParams as _GetTasksParams,
    GetTimetrackParams as _GetTimetrackParams,
    GetTimeZoneGroupsParams as _GetTimeZoneGroupsParams,
    GetTypesOfWorkParams as _GetTypesOfWorkParams,
    GetUsersParams as _GetUsersParams,
    GetUserScheduleParams as _GetUserScheduleParams,
    GetWorkflowStatusesParams as _GetWorkflowStatusesParams,
)


class ApiException(Exception):

    def __init__(self, code: int, err: ApiError, *args) -> None:
        super().__init__(*args)
        self.code = code
        self.err = err


T = TypeVar("T", bound=ApiModel)


class ApiClient:

    def __init__(self, url: str, auth: tuple[str, str] | AuthBase):
        self._url = url
        self._auth = auth

    def _get(
        self, path: str, params: Optional[dict] = None, headers: Optional[dict] = None
    ):
        res = requests.get(
            self._url + path, params=params, headers=headers, auth=self._auth
        )
        match res.status_code:
            case 200:
                return res.json()
            case _:
                err = ApiError.model_validate(res.json())
                raise ApiException(res.status_code, err)

    def _get_data(
        self,
        cls: Type[T],
        path: str,
        kw: Optional[_ParamsDict] = None,
    ):
        params = None
        if kw is not None:
            params = self.kw_to_params(kw)
        data = self._get(path, params=params)
        return cls.model_validate(data)

    # Customers

    def get_customer(self, cid: int):
        """Returns properties of customer with given id"""
        return self._get_data(CustomerWithAllowedAction, f"/customers/{cid}")

    def get_customers(self, /, **kw: Unpack[_GetCustomersParams]):
        """Returns list of customers according to provided query and user access rights"""
        return self._get_data(CustomerList, "/customers", kw)

    def get_customer_comments(self, cid: int, /, **kw: Unpack[_GetCommentsParams]):
        """Returns list of comments on the customer according to provided query
        and user access rights"""
        return self._get_data(CommentList, f"/customers/{cid}/comments", kw=kw)

    # Departments

    def get_department_by_id(self, did: int):
        """Returns properties of department with given id"""
        return self._get_data(Department, f"/departments/{did}")

    def get_departments(self, **kw: Unpack[_GetDepartmentsParams]):
        """Returns list of departments according to provided query"""
        return self._get_data(DepartmentList, "/departments", kw=kw)

    # Rest Hooks

    def get_hooks(self):
        """Returns list of available subscriptions"""
        data = self._get("/hooks")
        return [WithDetails.model_validate(d) for d in data]

    # Info

    def get_info(self):
        """Returns actiTIME instance info and settings"""
        return self._get_data(Info, "/info")

    # Leavetime

    def get_leavetime(self, **kw: Unpack[_GetLeavetimeParams]):
        """Returns leave time according to provided query. Leave time for single date is returned entirely."""
        return self._get_data(LeaveTimeList, "/leavetime", kw=kw)

    # Leave types

    def get_leavetype_by_id(self, uid: int):
        """Returns properties of leave type with given id"""
        return self._get_data(LeaveType, f"/leaveTypes/{uid}")

    def get_leavetypes(self, **kw: Unpack[_GetLeaveTypeParams]):
        """Returns list of leave types according to provided query"""
        return self._get_data(LeaveTypeList, "/leaveTypes", kw=kw)

    # Projects

    def get_project_by_id(self, uid: int):
        """Returns properties of project with given id"""
        return self._get_data(ProjectWithAllowedActions, f"/projects/{uid}")

    def get_project_comments(self, uid: int, /, **kw: Unpack[_GetCommentsParams]):
        """Returns list of comments on the project according to provided query and user access rights"""
        return self._get_data(CommentList, f"/projects/{uid}/comments", kw=kw)

    def get_ptojects(self, /, **kw: Unpack[_GetProjectsParams]):
        """Returns list of projects according to provided query and user access rights"""
        return self._get_data(ProjectList, "/projects", kw=kw)

    # Tasks

    def get_task_by_id(self, uid: int):
        """Returns properties of task with given id"""
        return self._get_data(TaskWithAllowedActions, f"/tasks/{uid}")

    def get_task_comments(self, uid: int, **kw: Unpack[_GetCommentsParams]):
        """Returns list of comments on the task according to provided query and user access rights"""
        return self._get_data(CommentList, f"/tasks/{uid}/comments", kw=kw)

    def get_tasks(self, **kw: Unpack[_GetTasksParams]):
        """Returns list of tasks according to provided query and user access rights"""
        return self._get_data(TaskList, "/tasks", kw=kw)

    # Timetrack

    def get_single_timetrack(self, user_id: int, date: DateTime, task_id: int):
        """Returns one time-track record"""
        return self._get_data(
            TimeTrackRecord, f"/timetrack/{user_id}/{serialize_date(date)}/{task_id}"
        )

    def get_timetrack(self, **kw: Unpack[_GetTimetrackParams]):
        """Returns several time-track records according to provided query."""
        return self._get_data(TimeTrackList, "/timetrack", kw=kw)

    # Time Zone Groups

    def get_time_zone_group_by_id(self, uid: int):
        """Returns properties of time zone group with given id"""
        return self._get_data(TimeZoneGroup, f"/timeZoneGroups/{uid}")

    def get_default_time_zone_group(self):
        """Returns default time zone group"""
        return self._get_data(TimeZoneGroup, "/timeZoneGroups/default")

    def get_time_zone_groups(self, **kw: Unpack[_GetTimeZoneGroupsParams]):
        """Returns list of time zone groups according to provided query"""
        return self._get_data(TimeZoneGroupList, "/timeZoneGroups", kw=kw)

    # Types of work

    def get_type_of_work_by_id(self, uid: int):
        """Returns properties of type of work with given id"""
        return self._get_data(TypeOfWork, f"/typesOfWork/{uid}")

    def get_default_type_of_work(self):
        """Returns default type of work"""
        return self._get_data(TypeOfWork, "/typesOfWork/default")

    def get_types_of_work(self, **kw: Unpack[_GetTypesOfWorkParams]):
        """Returns list of types of work according to provided query"""
        return self._get_data(TypeOfWorkList, "/typesOfWork", kw=kw)

    # Users

    def get_users(self, **kw: Unpack[_GetUsersParams]):
        """Returns list of users according to provided query"""
        return self._get_data(UserList, "/users", kw=kw)

    def get_user_by_id(self, uid: int):
        """Returns properties of user with given id"""
        return self._get_data(UserWithAllowedActions, f"/users/{uid}")

    def get_me(self):
        """Returns properties of authorized user"""
        return self._get_data(UserWithAllowedActions, "/users/me")

    def get_user_schedule(self, uid: int, **kw: Unpack[_GetUserScheduleParams]):
        """Returns user schedule for provided dates"""
        return self._get_data(UserSchedule, f"/users/{uid}/schedule", kw=kw)

    # Workflow Statuses

    def get_workflow_status_by_id(self, uid: int):
        """Returns properties of workflow status with given id"""
        return self._get_data(
            WorkflowStatusWithAllowedActions, f"/workflowStatuses/{uid}"
        )

    def get_workflow_statuses(self, **kw: Unpack[_GetWorkflowStatusesParams]):
        """Returns list of workflow statuses according to provided query"""
        return self._get_data(WorkflowStatusList, "/workflowStatuses", kw=kw)

    @staticmethod
    def kw_to_params(d: _ParamsDict):
        _params = dict()
        for k, v in d.items():
            _k = to_camel(k)
            if v is None:
                continue
            if isinstance(v, DateTime):
                _v = serialize_date(v)
            elif isinstance(v, Sequence) and not isinstance(v, str):
                _v = ",".join([str(e) for e in v])
            elif isinstance(v, Enum):
                _v = v.value
            else:
                _v = v
            _params[_k] = _v
        return _params
